from qgis.PyQt.QtCore import (QFile, QRegExp, QSize, Qt)
from qgis.PyQt.QtGui import (QIcon, QKeySequence, QStandardItem)
from qgis.PyQt.QtWidgets import (QApplication, QComboBox, QDialog, QFileDialog, QLabel, QLineEdit, QListWidget, QPushButton, QProgressBar, QProxyStyle, QStyle,
                                 QTextEdit, QTreeWidget, QTreeWidgetItem)
from qgis.PyQt.QtPrintSupport import (QPrintPreviewDialog, QPrinter)
from qgis.core import (Qgis, QgsApplication, QgsLayerTree, QgsProject, QgsRasterLayer)
from os.path import (basename, dirname, exists, isfile, join, splitext)
from sys import path as sysPath
from xml.dom.minidom import parseString


try:
    from osgeo.gdal import Open as gdalopen
    from osgeo.gdal import GetDriver as gdal_GetDriver
    from osgeo.gdal import GetDriverCount as gdal_GetDriverCount
    from osgeo.gdal import GetDriverByName as gdal_GetDriverByName
    from osgeo.gdal import (DMD_LONGNAME, DMD_MIMETYPE, DMD_EXTENSION)
    from osgeo.ogr import GetDriver as ogr_GetDriver
    from osgeo.ogr import GetDriverCount as ogr_GetDriverCount
except:
    from gdal import Open as gdalopen
    from gdal import GetDriver as gdal_GetDriver
    from gdal import GetDriverCount as gdal_GetDriverCount
    from gdal import GetDriverByName as gdal_GetDriverByName
    from gdal import (DMD_LONGNAME, DMD_MIMETYPE, DMD_EXTENSION)
    from ogr import GetDriver as ogr_GetDriver
    from ogr import GetDriverCount as ogr_GetDriverCount


#NEW 3.2.0
from math import (floor, log, pow)

def convertSize(size):
   size_name = ("", "K.", "M.", "G.", "T.", "P.", "E.", "Z.", "Y.")
   i = 0
   if size > 0 :
      i = int(floor(log(size, 1000)))
      p = pow(1000, i)
      s = round(size/p, 2)
   else :
      s = 0
   return '%s %s ' % (s, size_name[i])
    
##### ======= FUNCTIONS QTREEWIDGET ======== ####
def visitTreeElt(listItems, item, level, maxlevel):
  listItems.append([level, item.text(0), item.whatsThis(0), item.text(1)])
  for i in range(item.childCount()):
      if maxlevel < level+1 : maxlevel = level+1  
      maxlevel = visitTreeElt(listItems, item.child(i), level+1, maxlevel)
  return maxlevel    

def visitTreeList(tree) :
  listItems = []
  maxlevel = 0
  for i in range(tree.topLevelItemCount()) :
      maxlevel = visitTreeElt(listItems, tree.topLevelItem(i), 0, 1)
  return listItems, maxlevel
    
def legendSymbolItemChecked(node, ruleKey):
    isCheckedElt = False
    if Qgis.QGIS_VERSION_INT <= 31299 : isCheckedElt = node.layer().renderer().legendSymbolItemChecked(ruleKey)
    else :
          try : isCheckedElt = node.layer().featureRenderer().legendSymbolItemChecked(ruleKey)
          except : isCheckedElt = node.layer().renderer().legendSymbolItemChecked(ruleKey) #NEW AFTER QGIS 3.4
    return isCheckedElt

def conditionnalAddChild(treeWidget, parent, legendItem) :    
    if not parent : treeWidget.addTopLevelItem(legendItem) 
    else : parent.addChild(legendItem)

def createTreeWidgetItem(iconPath, icon, text):
    legendItem = QTreeWidgetItem()
    legendItem.setIcon(0, icon)
    legendItem.setText(0, text)
    legendItem.setWhatsThis(0, iconPath)
    return legendItem    
##### ======= END FUNCTIONS QTREEWIDGET ======== ####    

##### =======    FUNCTIONS LISTLAYERS   ======== #### 
def getListLayers(self, targetList, model, comboBox):
   root = QgsProject.instance().layerTreeRoot()
   targetList.clear()
   comboBox.clear()
   
   targetList[0], maxDim, dim = [], 0, 400

   i = 0
   if root != None :
       #NEW 3.2.0
       nameGroup = ""
       for layer in root.findLayers() : 
           if type(layer.layer()) == QgsRasterLayer :
              #NEW 3.2.0 - FOR GroupComboBox  
              if nameGroup != layer.parent().name():
                 nameGroup = layer.parent().name() 
                 try : comboBox.addGroup(model, nameGroup)
                 except : pass
              if not dicoHasKey(targetList, layer.layerId()):
                 dim = addItemInModel(self, targetList, model, layer)
                 maxDim = max(maxDim, dim)
   
   comboBox.setModelColumn(0)
   comboBox.show()
   comboBox.setCurrentIndex(-1)
   
   return maxDim

def addItemInModel(self, targetList, model, layer):
   item = QStandardItem(QIcon(getThemeIcon("raster.png")), "%s, [ %s ]" % (layer.name(), layer.layer().crs().authid()))
   item.setData("%s" % layer.layerId(), Qt.UserRole)
   font = QFont('Times New Roman', 9)
   font.setItalic(True)        
   item.setFont(font)   
   model.appendRow(item) 
   targetList[layer.layerId()] = []
   return len(layer.name())

def getComboLayer(self, comboBox, index):
   targetLayer = None
   if index != -1 :
       modelIndex = comboBox.model().index(index, 0)
       layerId = comboBox.model().data(modelIndex, Qt.UserRole)
       root = QgsProject.instance().layerTreeRoot()
       if root != None :
               for layer in root.findLayers() :
                   if layer.layerId() == layerId: targetLayer = layer.layer(); break;
   return targetLayer

def getComboLayerIndex(self, comboBox, text):
   model, index = comboBox.model(), 0
   for i in range(comboBox.count()):
       if model.item(i).text() == text :    
          index = i  
          break
   return index
##### =======  END FUNCTIONS LISTLAYERS ======== ####


def textForMenu(pattern, text, maxDim, shorcuts):
    return pattern % (text, shorcuts.ljust(maxDim-len(text)-len(shorcuts))) if pattern else text 

def fixeMaxDim(listItems):
    maxDim = 40
    if listItems != []:
       for item in listItems: maxDim = max(maxDim, len(item))
    return maxDim

def iff(a, b, c):
    if a: return b
    else: return c

def makeListFormatsExtensions(self, vector, raster, includeJoker):
    listFormatsExtensions, listFormatsExtensionsVector, listFormatsExtensionsAll = [], [], []

    for iDriver in range(ogr_GetDriverCount()):
        poDriver = ogr_GetDriver(iDriver)
        poValue = poDriver.GetName()
        driver = gdal_GetDriverByName("%s" % (poValue))
        listFormatsExtensionsVector = setEltListFormatsExtensions(self, listFormatsExtensionsVector, driver, poValue)
        

    for iDriver in range(gdal_GetDriverCount()):
        poDriver = gdal_GetDriver(iDriver)
        poValue = poDriver.ShortName
        driver = gdal_GetDriverByName("%s" % (poValue))
        listFormatsExtensionsAll = setEltListFormatsExtensions(self, listFormatsExtensionsAll, driver, poValue)

    if vector and raster : listFormatsExtensions =  listFormatsExtensionsAll
    else :
        if vector : listFormatsExtensions = listFormatsExtensionsVector
        elif raster :
            listFormatsExtensions = [value for value in listFormatsExtensionsAll if value not in listFormatsExtensionsVector] 
            if not "jp2" in listFormatsExtensions : listFormatsExtensions.append('jp2')
            
    if includeJoker : listFormatsExtensions.append('*')
    return sorted(listFormatsExtensions)

def setEltListFormatsExtensions(self, listFormatsExtensions, driver, poValue):
       if driver :
            poDriverMetadata = driver.GetMetadata()
            if dicoHasKey(poDriverMetadata, DMD_EXTENSION):
                if poDriverMetadata[DMD_EXTENSION]!= '' and not poDriverMetadata[DMD_EXTENSION] in  listFormatsExtensions : listFormatsExtensions.append(poDriverMetadata[DMD_EXTENSION])
       else :
           if poValue != '' and not poValue in  listFormatsExtensions : listFormatsExtensions.append[poValue]
       return listFormatsExtensions


def getExtension(fileName):
    from os.path import (basename, splitext)
    textension = splitext(fileName)
    extension = textension[len(textension)-1].lower()
    return extension

def windowIcon(qtWindow, icon):
    if not qtWindow : return
    icon = getThemeIcon(icon)
    if icon != "" : qtWindow.setWindowIcon(QIcon(icon)) 

def changeButtonIcon(self, button, icon, iconWidth, iconHeight):
    icon = getThemeIcon(icon)
    button.setIcon(QIcon(icon))

def getThemeIcon(theName):
    dirs = list(sysPath)
    dirs.insert(0, "%s%s" % (cleanPath(dirname(__file__)),"icons/"))
    dirs.insert(0, "%s%s" % (cleanPath(QgsApplication.pluginPath()),"autoLoaderRaster/icons/"))
    for dir in dirs :
        myPath = cleanPath(dir)
        myIconPath = "%s%s" % (myPath, theName)
        if QFile.exists(myIconPath): return myIconPath
    return ""
    
def cleanPath(path):
    path = "%s" % (path)
    a = len(path)
    letter = "/"
    b = path.rfind(letter, 0, a)
    return ("%s/" % (path)) if a != b else path 
    
def fixeLabelsFileDialog(self, dlgFile, context, hasName):
    dlgFile.setLabelText(QFileDialog.Accept, QApplication.translate("autoLoaderRasterPlugin","Open")) if context == 0 else dlgFile.setLabelText(QFileDialog.Accept, QApplication.translate("autoLoaderRasterPlugin","Save"))
    dlgFile.setLabelText(QFileDialog.Reject, QApplication.translate("autoLoaderRasterPlugin","Cancel"))        
    dlgFile.setLabelText(QFileDialog.FileName, "%s : " % (QApplication.translate("autoLoaderRasterPlugin","Name"))) if hasName else dlgFile.setLabelText(QFileDialog.FileName, QApplication.translate("autoLoaderRasterPlugin","Folder : "))
    dlgFile.setLabelText(QFileDialog.FileType, "%s : " % QApplication.translate("autoLoaderRasterPlugin","Filters"))
    dlgFile.setLabelText(QFileDialog.LookIn,   "%s : " % QApplication.translate("autoLoaderRasterPlugin","Folder"))

def dicoHasKey(dico, key):
    cond = False
    try : cond = dico.has_key(key)
    except : cond = (key in dico)
    return cond

def prettifyXml(xml):
    return parseString(xml).toprettyxml()


class textEditFolder(QTextEdit):
    def __init__(self, * args):
        QTextEdit.__init__(self, * args)

    def initTextEdit(self, SizeW, SizeH, posX, posY, name, isGeom, undoRedo, isReadOnly, isRichText, isSizable):
        if not isSizable:  self.setMaximumSize(SizeW, SizeH)
        if isGeom: self.setGeometry(posX, posY, SizeW, SizeH)
        self.setUndoRedoEnabled(undoRedo)
        self.setReadOnly(isReadOnly)
        self.setAcceptRichText(isRichText)
        self.setObjectName(name)
        self.setAccessibleName(name)
        self.setAcceptDrops(True)

class editor(QTextEdit):
    def __init__(self, parent, withProgressBar):
        super(editor, self).__init__(parent)
        self.setWindowTitle("QTextEdit Standard Output Redirection")
        self.setStyleSheet("color: white; background-color: black;")
        self.setAcceptRichText(True)
        self.setObjectName(parent.objectName())
        self.parent = parent
        self.progressBar = None

        if withProgressBar :
            self.progressBar = progressBar(self)
            self.progressBar.initTextEdit(200, 25, 200, 140, "progressBar", True, 0, True, Qt.AlignCenter, 0)
            self.progressBar.setStyleSheet("progressBar {border: 2px solid grey; border-radius: 5px; text-align: center;}"
                                            "progressBar::chunk {background-color: #6C96C6; width: 20px;}")


class funnyPushButton(QPushButton):
     def __init__(self, *args): super(funnyPushButton, self).__init__(*args)
            
     def initPushButton(self, sizeWidth, sizeHeight, coordX, coordY, name, text, toolTip, isGeom, icon, iconWidth, iconHeight, isStyleSheeted):
            self.setMinimumSize(sizeWidth, sizeHeight)
            self.setMaximumSize(sizeWidth, sizeHeight)
            self.iconWidth = iconWidth
            self.iconHeight = iconHeight
            self.selfFocused = False
            self.subMenuVisble = False
            if isGeom : self.setGeometry(coordX, coordY, sizeWidth, sizeHeight)
            if icon != "" :
               self.setIcon(QIcon(icon))
               self.setIconSize(QSize(self.iconWidth , self.iconHeight))
            self.setToolTip(toolTip)
            if isStyleSheeted : self.setStyleSheet(" QPushButton {border-width: 0px; border-radius: 10px;  border-color: beige;}")
            self.setObjectName(name)
            if text != "" : self.setText(text)

     def enterEvent(self, event):
          if not self.selfFocused :
             self.setIconSize(QSize(self.iconWidth *0.8, self.iconHeight *0.8))
             self.selfFocused = True
          
     def leaveEvent(self, event):
        if self.subMenuVisble == False : self.restoreButton()
     
     def restoreButton(self):
           self.selfFocused = False    
           self.setIconSize(QSize(self.iconWidth, self.iconHeight))
           
      
class labeller(QLabel):
     def __init__(self, name, text, styleSheet, alignment, *args):
             super(labeller, self).__init__(*args)
             self.setObjectName(name)
             self.setText(text)
             self.setMyStyleSheet(styleSheet)
             """
             if styleSheet == "special": self.setStyleSheet("QLabel {display: block;  border: 2px solid gray;  border-radius: 10px; text-align: center;  line-height: 10px; background-color: #C0C0C0; border: 1px solid black; color: black}")
             elif styleSheet == "infos": self.setStyleSheet("QLabel {font-size:10px; background-color:#C0C0C0; border: 1px solid #e1e1e1; padding: 4px ; color: black}")
             elif styleSheet == "coords": self.setStyleSheet("QLabel {  color : white; font-size: 11px; background-color: rgba(93, 93, 93, 125);}")
             """
             self.setAlignment(alignment)

     #NEW 3.0.0
     def setMyStyleSheet(self, style):
         styleSheet = {"special": "QLabel {display: block;  border: 2px solid gray;  border-radius: 10px; text-align: center;  line-height: 10px; background-color: #C0C0C0; border: 1px solid black; color: black}",
                       "infos": "QLabel {font-size:10px; background-color:#C0C0C0; border: 1px solid #e1e1e1; padding: 4px ; color: black}",
                       "normal": "QLabel {font-size:10px; background-color:#C0C0C0; border: 1px solid #e1e1e1; padding: 4px ; color: black}",
                       "coords": 'QLabel {color : white; font-size: 11px; background-color: rgba(93, 93, 93, 125)}',
                       "error" : "QLabel {display: block;  border: 2px solid gray;  border-radius: 10px; text-align: center;  line-height: 10px; background-color: #CC0000; border: 1px solid red; color: white}"}

         styleCss = styleSheet[style]
         self.setStyleSheet(styleCss)
      
         
         

class myPushButton(QPushButton):
     def __init__(self, *args): super(myPushButton, self).__init__(*args)
            
     def initPushButton(self, sizeWidth, sizeHeight, coordX, coordY, name, text, toolTip, isGeom, icon, iconWidth, iconHeight, isStyleSheeted):
            self.setMinimumSize(sizeWidth, sizeHeight)
            self.setMaximumSize(sizeWidth, sizeHeight)
            if isGeom : self.setGeometry(coordX, coordY, sizeWidth, sizeHeight)
            if icon != "" :
               self.setIcon(QIcon(icon))
               self.setIconSize(QSize(iconWidth, iconHeight))
            self.setToolTip(toolTip)
            if isStyleSheeted : self.setStyleSheet("QPushButton {border: none;}")
            self.setObjectName(name)
            if text != "" : self.setText(text)

class progressBar(QProgressBar):
    def __init__(self, * args):
        QProgressBar.__init__(self, * args)

    def initTextEdit(self, width, heigth, posX, posY, name, isGeom, initValue, viewText, alignement, isStyleSheet):
        self.setMinimumSize(width, heigth)
        self.setMaximumSize(width, heigth)
        if isGeom: self.setGeometry(posX, posY, width, heigth)
        self.setProperty("value", initValue)
        self.setAlignment(alignement)
        self.setTextVisible(viewText)
        self.setObjectName(name)
        self.setAccessibleName(name)
        if isStyleSheet == 0:
            self.setStyleSheet(
                               "progressBar {border: 2px solid grey; border-radius: 5px; text-align: center;}"
                               "progressBar::chunk {background-color: #6C96C6; width: 10px; margin: 0.5px;}"
                               )            


class printerDlg(object):
      def __init__(self, parent, obj, title, orientation):
                    
              printer = QPrinter() 
              printer.setPageSize(QPrinter.A4)
              printer.setOrientation(orientation)
              printer.setPageMargins(5, 10, 5, 10, QPrinter.Millimeter)
              printer.setOrientation(orientation)
              printer.setOutputFormat(QPrinter.NativeFormat)
              self = QPrintPreviewDialog(printer)
              windowIcon(self, "print.png")
              self.setWindowTitle(title)
              self.setWindowFlags(Qt.WindowMaximizeButtonHint | Qt.WindowStaysOnTopHint | Qt.WindowCloseButtonHint)
              self.paintRequested.connect(obj.print_)
              self.exec_()


class saveDlg(QDialog):
    def __init__(self, parent, initDir):
        QDialog.__init__(self)
        self.parent = parent
        self.initDir = initDir
        self.setupUi()

    def setupUi(self):
        self.setObjectName("saveDlg")
        self.maxWidth, self.maxheight = 620, 200 
        self.setMinimumSize(self.maxWidth,self.maxheight)
        self.setMaximumSize(self.maxWidth,self.maxheight)        
        self.resize(self.maxWidth, self.maxheight) 
        windowIcon(self, "checkdoubleglasses.png")

        self.setWindowTitle("%s ..." % QApplication.translate("autoLoaderRasterPlugin", "Save maps as images"))

        self.styleTextEdit = """
                    color: black; background-color: rgba(255, 255, 125, 50);
                    border-left-width: 1px;
                    border-left-color: black;
                    border-left-style: solid;
                    border-right-color: black;
                    border-right-width: 1px;
                    border-right-style: solid;
                    border-right-radius: 3px;"""

        self.styleTextEditDisaBled  = """
                    color: black; background-color: rgba(125, 125, 125, 50);
                    border-left-width: 1px;
                    border-left-color: black;
                    border-left-style: solid;
                    border-right-color: black;
                    border-right-width: 1px;
                    border-right-style: solid;
                    border-right-radius: 3px;"""

        self.authorizedMime = ["*.*"]


        icon = getThemeIcon("savefile.png")
        toolTip = QApplication.translate("autoLoaderRasterPlugin","Save maps as images")
        self.saveButton = funnyPushButton(self)
        self.saveButton.initPushButton(24, 24, self.width()-160, self.height()-30, "saveButton", "", toolTip, True, icon, 24, 24, True)
        self.saveButton.setShortcut(QKeySequence("Ctrl+S"))

        icon = getThemeIcon("close.png")
        toolTip = QApplication.translate("autoLoaderRasterPlugin","Close")
        self.closeButton = QPushButton(self)
        self.closeButton.setObjectName("closeButton")
        self.closeButton.setText("&%s" % QApplication.translate("QSphere", "Close"))
        self.closeButton.setGeometry(self.width()-120, self.height()-30, 100, 25)


        self.lblNameFirstImage = QLabel(self)
        self.lblNameFirstImage.setAlignment(Qt.AlignRight)
        self.lblNameFirstImage.setText("%s %s 1 :" % (QApplication.translate("autoLoaderRasterPlugin","Name"), QApplication.translate("autoLoaderRasterPlugin", "MapCanvas").lower()))
        self.lblNameFirstImage.setGeometry(20, 20, 120, 25)


        self.listExtension = QComboBox(self)
        self.listExtension.setGeometry(520, 20, 60, 25)
        self.listExtension.setEditable(False) 
        self.listExtension.setEnabled(True)
        self.listExtension.setObjectName("listExtension")
        #listExtensions = makeListFormatsExtensions(self, False, True, False)
        listExtensions = ["bmp", "jpg", "png", "tif"]
        self.listExtension.addItems(listExtensions)

        if "png" in listExtensions : self.listExtension.setCurrentIndex( listExtensions.index("png"))
                

        self.nameFirstImage = lineEditNameFile(self)
        self.nameFirstImage.initCtrlWidget(self.saveButton, self.listExtension)
        self.nameFirstImage.setGeometry(150, 20, 360, 25)
        self.nameFirstImage.setObjectName("nameFirstImage")
        self.nameFirstImage.setStyleSheet(self.styleTextEdit)
        self.nameFirstImage.setText("mapviewer1.png")
        self.nameFirstImage.textChanged.connect(self.nameFirstImage.checkRegExp)
        self.nameFirstImage.checkRegExp()

        self.lblNameSecondImage = QLabel(self)
        self.lblNameSecondImage.setAlignment(Qt.AlignRight)
        self.lblNameSecondImage.setText("%s %s 2 :" % (QApplication.translate("autoLoaderRasterPlugin","Name"), QApplication.translate("autoLoaderRasterPlugin", "MapCanvas").lower()))
        self.lblNameSecondImage.setGeometry(20, 50, 120, 25)


        self.nameSecondImage = lineEditNameFile(self)
        self.nameSecondImage.initCtrlWidget(self.saveButton, self.listExtension)
        self.nameSecondImage.setGeometry(150, 50, 360, 25)
        self.nameSecondImage.setObjectName("nameSecondImage")
        self.nameSecondImage.setStyleSheet(self.styleTextEdit)
        self.nameSecondImage.setText("mapviewer2.png")
        self.nameSecondImage.textChanged.connect(self.nameSecondImage.checkRegExp)
        self.nameSecondImage.checkRegExp()
        
        self.lblFolder = QLabel(self)
        self.lblFolder.setText("%s :" % QApplication.translate("autoLoaderRasterPlugin","Folder"))
        self.lblFolder.setAlignment(Qt.AlignRight)
        self.lblFolder.setGeometry(20, 80, 120, 25)
 
        self.pathToSaveImage = textEditFolder(self)
        self.pathToSaveImage.initTextEdit(360, 40, 150, 80,  "pathToSaveImage", True, False, True, False, True)
        self.pathToSaveImage.setStyleSheet(self.styleTextEditDisaBled)
        self.pathToSaveImage.setText(self.initDir)

        self.getFolderButton = funnyPushButton(self)
        self.getFolderButton.initPushButton(28, 28, 520, 80, "getFolderButton", "", QApplication.translate("autoLoaderRasterPlugin","Select a folder"), True, getThemeIcon("folder.png"), 28, 28, True)

        self.getFolderButton.clicked.connect(self.getFolder)

        self.listExtension.currentTextChanged.connect(self.updateExtension)
        self.saveButton.clicked.connect(self.createFilesImages)
        self.closeButton.clicked.connect(self.close)

    def updateExtension(self):
        widgets = (self.nameFirstImage, self.nameSecondImage)
        for widget in widgets :
            if widget.text().find(".") == 0 : widget.setText("%s.%s" % (widget.text(), self.listExtension.currentText()))
            else :
               extension = widget.text().split(".")[-1]
               widget.setText("%s%s" % (widget.text()[:-3], self.listExtension.currentText()))
            widget.checkRegExp()   
           
        

    def closeEvent(self, event):
        self.parent.initDir = self.initDir

    def getFolder(self):
        if self.sender() == None: return
        if self.sender() == self.getFolderButton : obj = self.pathToSaveImage
        else: return

        windowIcon(self, "folder.png")

        self.initDir = dirname(__file__) if obj.toPlainText() == "" else dirname(obj.toPlainText())
        title = QApplication.translate("autoLoaderRasterPlugin", "Select the folder")

        fileDialog = QFileDialog(self, title)
        fileDialog.setDirectory(self.initDir)
        fileDialog.setFileMode(QFileDialog.Directory)
        fileDialog.setOption(QFileDialog.ShowDirsOnly, False)
        fileDialog.setOption(QFileDialog.DontUseNativeDialog)
        fileDialog.setNameFilters(self.authorizedMime)
        fileDialog.selectNameFilter(self.authorizedMime[0])

        fixeLabelsFileDialog(self, fileDialog, 0, False) 

        if fileDialog.exec_():
            inputDir = fileDialog.selectedFiles()[0]
            obj.setPlainText(cleanPath(inputDir.replace("\\", "/")))
            self.initDir = obj.toPlainText()

        
    def createFilesImages(self):

        folder = self.pathToSaveImage.toPlainText()
        fileName1 = self.nameFirstImage.text()
        fileName2 = self.nameSecondImage.text()
        extension = self.listExtension.currentText().upper()

        if folder == "" : return
        if fileName1 == fileName2 : return

        fileName = join(folder, fileName1) 
        self.parent.mapViewerLayer.saveAsImage(fileName, None, extension)

        fileNameDouble = join(folder , fileName2) 
        self.parent.mapViewerLayerDouble.saveAsImage(fileNameDouble, None, extension)

        return fileName, fileNameDouble


class lineEditNameFile(QLineEdit):
      def __init__(self, *args):
          super(lineEditNameFile, self).__init__(*args)
          self.regex = QRegExp(r"^[\w\-\. ]+$", Qt.CaseSensitive)
          
      def initCtrlWidget(self, widgetCtrl, extension):
          self.widgetCtrl = widgetCtrl
          self.extension = extension

      def checkRegExp(self):
          self.setStyleSheet("background-color:red;")
          self.widgetCtrl.setEnabled(False)
          cond = self.regex.exactMatch(self.text())
          if self.text().find(".") == 0 : return
          extension = self.text().split(".")[-1]
          if cond and extension.lower() == self.extension.currentText().lower() : #"png" :
             self.setStyleSheet("background-color:#AEEE00;")
             self.widgetCtrl.setEnabled(True)
             
          
#NEW 3.0.0
class myListWidget(QListWidget):
    def __init__(self, parent=None):
        QListWidget.__init__(self, parent)
        self.parent = parent.parent()

    def dropEvent(self, event):
        super(myListWidget, self).dropEvent(event)
        self.parent.updateListFiles()
        self.parent.updateMapViewerWithGrid()

class textEdit(QTextEdit):
      def __init__(self, parent, *args):
          super(textEdit, self).__init__(*args)
          self.parent = parent
          
      def focusOutEvent(self, event): self.parent.updateInfo(self)


class myMenuStyle(QProxyStyle):
    pass
    def pixelMetric(self, QStyle_PixelMetric, option=None, widget=None):
        if QStyle_PixelMetric == QStyle.PM_SmallIconSize: return  20 
        else : return QProxyStyle.pixelMetric(self, QStyle_PixelMetric, option, widget)



class wizardData(QDialog):
    def __init__(self, parent):
        QDialog.__init__(self)
        self.parent = parent
        self.setupUi()
    
    def setupUi(self):
        self.setObjectName("ui_wizardData")
        self.maxWidth, self.maxheight = 400, 350 
        self.setMinimumSize(self.maxWidth, self.maxheight)
        self.setMaximumSize(self.maxWidth, self.maxheight)
        self.resize(self.maxWidth, self.maxheight) 
        windowIcon(self, "bggroup.png")

        self.setWindowTitle(QApplication.translate("autoLoaderRasterPlugin", "Select group raster"))


        self.groupsViewer = None
        self.groupIconPath = getThemeIcon("group.png").replace("\\","/")
        self.groupIcon = QIcon(self.groupIconPath)

        self.applyButton = QPushButton(self)
        self.applyButton.setObjectName("applyButton")
        self.applyButton.setText(QApplication.translate("autoLoaderRasterPlugin", "Validate"))

        self.closeButton = QPushButton(self)
        self.closeButton.setObjectName("closeButton")
        self.closeButton.setText(QApplication.translate("autoLoaderRasterPlugin", "Close"))

        self.closeButton.clicked.connect(self.close)

        sender = self.parent.sender()        
        
        self.getGroupsViewer(True)
        if self.groupsViewer :
           self.identifyGroup(self.parent.groupRaster.toPlainText())  
           self.groupsViewer.itemDoubleClicked.connect(self.setValueForGroupRaster)
           self.applyButton.clicked.connect(self.setValueForGroupRaster)
    



    def getGroupsViewer(self, getGroups):
        self.groupsViewer = QTreeWidget(self)
        self.groupsViewer.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.groupsViewer.setObjectName("groupsViewer")
        self.groupsViewer.headerItem().setText(0, "1")
        self.groupsViewer.header().setVisible(False)
        self.groupsViewer.header().setDefaultSectionSize(200)
        self.groupsViewer.setStyleSheet("QTreeWidget { font-size:12px; color: black; background-color: rgba(255, 255, 125, 50); border-left-width: 1px; border-left-color: black;  border-left-style: solid;  border-right-color: black; border-right-width: 1px; border-right-style: solid; border-right-radius: 3px;}"
                                       "QTreeWidget::item:selected {background-color: #5D5D5D; color: rgb(255,255,255); }")
        self.groupsViewer.setVisible(True)
        self.groupsViewer.setItemsExpandable(True)
        self.groupsViewer.setExpandsOnDoubleClick(True)
        self.groupsViewer.setContextMenuPolicy(Qt.NoContextMenu)

        self.populateGroupsViewer(getGroups)


    def identifyGroup(self, txt) :
        if txt != "" :
           items = self.groupsViewer.findItems(txt, Qt.MatchContains|Qt.MatchRecursive, 0) 
           if items :
              if items[0] : self.groupsViewer.setCurrentItem(items[0])

    
    def setValueForGroupRaster(self):
        self.parent.groupRaster.setPlainText(self.groupsViewer.currentItem().text(0))
        self.parent.updateInfo(self.parent.groupRaster)
        self.close()

    def introspectNodes(self, group, nodes, parent, getGroups):
        for node in nodes:
                if QgsLayerTree.isGroup(node): 
                   legendItem = createTreeWidgetItem(self.groupIconPath, self.groupIcon, node.name())
                   conditionnalAddChild(self.groupsViewer, parent, legendItem)
                   legendItem.setExpanded(node.isExpanded())
                   self.introspectNodes(node, node.children(), legendItem, getGroups)        


    def populateGroupsViewer(self, getGroups):
        if not self.groupsViewer : return
        self.groupsViewer.clear()
        root = QgsProject.instance().layerTreeRoot() 
        if getGroups : self.introspectNodes(root, root.children(), None, getGroups)
        else :
            listLayers = QgsProject.instance().mapLayers()
            dictLayers = {}
            for key in listLayers :
                layer = listLayers[key]
                if not layer.name() in dictLayers.keys() : dictLayers[layer.name()]=key
            
            iconTypes = {0: (self.pointIcon, self.pointIconPath) , 1 : (self.lineIcon, self.lineIconPath),  2: (self.polygonIcon, self.polygonIconPath)}
        
            for layerName in sorted(dictLayers) :
                key = dictLayers[layerName]
                layer = listLayers[key]
                if layer.name()!= "tempo" and type(layer) == QgsVectorLayer :
                   
                   typeGeom = layer.geometryType()
                   if dicoHasKey(iconTypes, typeGeom) : icon, iconPath = iconTypes[typeGeom][0], iconTypes[typeGeom][1]
                   else  :  icon, iconPath = self.unknowIcon, self.unknowIconPath
                    
                   legendItem = createTreeWidgetItem(iconPath, icon, layer.name())
                   conditionnalAddChild(self.groupsViewer, None, legendItem)    

    def close(self): self.reject()

    def resizeEvent(self,ev):
        self.groupsViewer.setGeometry(10, 10, self.width()-20, self.height()-60)
        self.applyButton.setGeometry(self.width()-230, self.height()-35, 100, 25)
        self.closeButton.setGeometry(self.width()-120, self.height()-35, 100, 25)
        

#NEW 3.2.0
from qgis.PyQt.QtWidgets import QStyledItemDelegate
from qgis.PyQt.QtGui import (QBrush, QColor, QFont, QStandardItemModel)
        
class GroupDelegate(QStyledItemDelegate):
    def initStyleOption(self, option, index):
        super(GroupDelegate, self).initStyleOption(option, index)
        if not index.data(Qt.UserRole):
            option.text = "   %s" % option.text


class GroupItem(QStandardItem):
    def __init__(self, text):
        super(GroupItem, self).__init__(text)
        self.setData(True, Qt.UserRole)
        self._number_of_childrens = 0
        font = QFont('Times New Roman', 10)
        font.setBold(True)
        font.setUnderline(True)
        font.setOverline(True)
        self.setFont(font)
        self.setFlags(self.flags() & ~Qt.ItemIsSelectable)

    def addChild(self, text):
        it = QStandardItem(text)        
        it.setData(False, Qt.UserRole)
        self._number_of_childrens += 1
        self.model().insertRow(self.row() + self._number_of_childrens, it)
        return it


class GroupComboBox(QComboBox):
    def __init__(self, parent=None):
        super(GroupComboBox, self).__init__(parent)
        self.setModel(QStandardItemModel(self))
        delegate = GroupDelegate(self)
        self.setItemDelegate(delegate)
        self.setIconSize(QSize(24, 24))
        self.iconGroup = QIcon(getThemeIcon("group.png"))
        self.iconChild = QIcon(getThemeIcon("raster.png"))

    def addGroup(self, model, text):
        it = GroupItem(text)
        it.setBackground(QBrush(QColor(25, 230, 240, 200)))
        model.appendRow(it)
        self.setItemIcon(self.count()-1, self.iconGroup)
        return it

    def addChild(self, text):
        it = QStandardItem(text)
        it.setData(True, Qt.UserRole)
        self.model().appendRow(it)
        self.setItemIcon(self.count()-1, self.iconChild)
        return it        
